=begin rdoc
ruby

Ulrich Mutze, www.ulrichmutze.de 

2008-12-09

This is free software. It comes without any warranty.
Please contact the author to report problems or to
ask questions.

The intention is that a statement
  require 'appmath_basics.rb'
allows us to use in the following a programming style in which the concept
of real numbers is handled in a flexibel manner so that switching from
standard type Float to arbitrary precision type R can be achieved
by a single statement of the type
  R.dig = -1 ( for using Float)
or 
  R.dig = 100 ( for using R with 100 decimal places ) 
All modules and classes to be defined furtheron within the module
  AppMath
will require appmath_basics.rb and will use real numbers in the manner
indicated above.
=end 

require 'rnum'
require 'float_ext'

# A Module which collects concepts of applied mathematics and of physics.
module AppMath 

# A Module which collects simple utility functions.
module Basics

# Returns the larger of the two numbers a and b.
def Basics.sup(a, b)
  a <= b ? b : a
end 

# Returns the smaller of the two numbers a and b.
def Basics.inf(a, b)
  a < b ? a : b
end 

# Returns the sign of a as a real number.
def Basics.sign(a)
  if a > R.c0
    R.c1
  elsif a < R.c0
    -R.c1
  else
    R.c0
  end
end

# Returns 'a with the sign of b'. Programming idiom frequently used in
# the 'Numerical Recipes in C' by Press et al.
def Basics.sign2(a, b)
  b >= R.c0 ? a.abs : -a.abs
end

# Returns the nearest number to x, which up to sign and power of 10
# is one of the numbers in the array c in the function
# body. For instance,
#   c = [1,2,5,7.5,10]
# In a sense, here cutting is a more drastic version of rounding.
def Basics.cut(x)
  fail "x is not a number" if x.nan?
  fail "x is not finite" if x.infinite?
  c = [R.c1,R.c2,R.c5,R.c(7.5),R.c10]
  return R.c0 if x.zero? 
  s = Basics.sign(x)
  y1 = x.abs
  ylog = y1.log10
  ylogInt = ylog.floor
  yMantisse = ylog - ylogInt
  y = R.c10 ** yMantisse
  fail "Basics.cut(): !(1 <= y)" if !(R.c1 <= y)
  fail "Basics.cut(): !(y < 10)" if !(y < R.c10)
  i = 0
  while y > c[i] 
    i += 1
  end
  cd = c.length
  fail "unexpected failure in function Basics.cut" if i < 0 || i >= cd
  yu = c[i] # yu is the smallest c[k] which is larger or equal to y
  if i==0
    yf = yu
  else
    fail "unexpected failure in function Basics.cut" if (i-1) < 0 || (i-1) >= cd
    yl = c[i-1]
    yf = yu-y <y -yl ? yu : yl # now yf is the c[k] which is nearest to y
  end
  return yf.ldexp(ylogInt) * s # sign and exponent are restored
end  

end # Basics

end # module AppMath