=begin rdoc
ruby

Ulrich Mutze, www.ulrichmutze.de

2008-12-07

This is free software. It comes without any warranty.
Please contact the author to report problems or to
ask questions.
=end 

# Usage:
#   require 'float_ext.rb'
#   x = 1.37; y = 0.123
#   phi = x.arg(y)
#   r = x.hypot(y)
#   dx = r * phi.cos - x
#   dy = r * phi.sin - y
#
# Defining all functons of module Math as methods of class Float. 
# The meaning of functions frexp and ldexp has been changed from 
# referring to exponent 2 to exponent 10.
# Notice that code that worked without a statement
#   require 'float_ext.rb'
# will work in an identical manner with this statement added. 
# Only if this is true, extending the functionaliy of standard types is
# considered acceptable.
# A few functions, not to be found in module Math have been added.
# These are methods which are defined in R and thus should be available
# in Float in order to make R and Float strictly replacable in all code
# that loads float_ext.rb and rnum.rb.
# These new functions are:
#   inv, pseudo_inv, conj, cot, coth, acot, acoth, arg,  
#   clone, dis, integer?, real?, complex? 
# Notice also, that for R.dig = -1, the real-number generating functions
#   R.ran, R.tob, R.c, R.i, R.pi, R.e
#   R.c0, R.c1, ..., R.c10, R.i2, ... R.i10
# that normally return R-numbers are forced to return Floats so that it
# is simple to write programs in such a manner that all real numbers
# switch their type from Float to R and vice versa. This is one mode 
# more to use Float. Note that setting R.dig = 0 also uses Float, but
# does so in a hidden manner by basing the representation of R-numbers
# internally on Float. If an application has to switch between Float
# encoding of numbers and BigDecimal encoding of numbers one better 
# uses R for all real numbers both and changes the representation by
# calling the function R.dig=.
# The class R is coded in a manner that it makes no use of the present 
# extension of Float, although this would have allowed some code
# reduction.

class Float

  def sqrt; Math.sqrt(self); end
  
  # x.hypot(y) is an efficient and accurate representation of the
  # square root of x*x + y*y. 
  def hypot(y); Math.hypot(self,y); end
  
  # It is convenient to have inversion (the multiplicative analogon
  # of the unary - operation) as a member function.
  def inv; 1.0/self; end
  
  # pseudo inverse, which always exists
  def pseudo_inv; zero? ? 0.0 : 1.0/self; end
  
  # (Complex) conjugation, no effect on real numbers.
  # Supports the unified treatment of real and complex numbers.
  def conj; self; end

  def exp; Math.exp(self); end
  def log; Math.log(self); end
  def log10; Math.log10(self); end

  def sin; Math.sin(self); end
  def sinh; Math.sinh(self); end
  def asin; Math.asin(self); end
  def asinh; Math.asinh(self); end
  
  def cos; Math.cos(self); end
  def cosh; Math.cosh(self); end
  def acos; Math.acos(self); end
  def acosh; Math.acosh(self); end
  
  def tan; Math.tan(self); end
  def tanh; Math.tanh(self); end
  def atan; Math.atan(self); end
  def atan2(x); Math.atan2(self,x); end
  
  # The name 'arg' stands for 'agument', which is the name for the 
  # polar angle of a complex number preferred in the mahematical
  # literature. So for a complex number z = u + iv we have
  #   argument(z) = u.arg(v)
  # Notice also the representation of the absolute value of z:
  #   |z| = u.hypot(v)
  # The functions hypot and arg (or atan2) thus prepare the introduction
  # of complex numbers.
  def arg(y); Math.atan2(y,self); end
  
  def atanh; Math.atanh(self); end
  
  # All functions having 'cot' in their name deal with the trigonometric
  # or hyperbolic cotangent. These functions are so directly related to 
  # the corresponding 'tan'-functions that they are not included in the
  # Math and BigMath modules. It is sometimes useful to have them, though.
  def cot; Math.cos(self)/Math.sin(self); end
  def coth; Math.cosh(self)/Math.sinh(self); end
  def acot; Math.atan(1./self); end 
  def acoth; 0.5 * Math.log( ((self + 1.0)/(self - 1.0)).abs ) end
  
  # error function
  def erf; Math.erf(self); end
  
  # complemetary error function
  def erfc; Math.erfc(self) ; end
  
  # Warning! x.frexp differs from Math.frexp(x).
  # We need exponent 10 not 2,
  # so we don't use: 
  #   def frexp; Math.frexp(self);end
  def frexp 
  #  puts "arg of frex="+to_s
    if zero?
      [ 0.0, 0]
    elsif self > 0.0
      y = Math.log10(self)
      yf = y.floor
      yfrac = y - yf
      [ 10.0 ** yfrac, yf.to_i]
    else
      y = Math.log10(-self)
      yf = y.floor
      yfrac = y - yf
      [ - 10.0 ** yfrac, yf.to_i]
    end
  end
  
  # Warning! x.ldexp differs from Math.ldexp(x).
  # We need exponent 10 not 2,
  # so we don't use: 
  #   def ldexp(n); Math.ldexp(self,n);end
  def ldexp(n); self * (10.0 ** n.to_i); end 

  # One should have this!
  def clone
    res = 0.0; res += self; res
  end
 
# For all Float x, y we have 0 <= x.dis(y) <= 1 and x.dis(x) = 0
# It is a kind of relative distance which should be in the order
# of magnitude of the smalles positive representable number
# if x and y are known to differ only by numerical noise. 
  def dis(x)
    xf = x.to_f
    a = abs
    b = xf.abs
    d = (self - xf).abs
    s = a + b
    return 0.0 if s.zero?
    d1 = d/s
    d < d1 ? d : d1
  end
  
# Since Float is not Fixnum or Bignum we return 'false'. In scientific
# computation there may be the need to use various types of 'real number
# types' but there should be always a clear-cut distinction between
# integer types and real types.
  def integer?; false; end  

# Although there may be technical variants in representing real numbers,
# these all should answer this question with 'yes' since they all
# model mathematical real numbers. 
  def real?; true; end

# Not complex, since no second dimesion (filling a plane) is provided. 
  def complex?; false; end

# Providing the neutral element of addition via a method.
  def to_0; 0.0; end

# Providing the neutral element of multiplication via a method.
  def to_1; 1.0; end
  
 # Print. Output to console, together with a name which is given by the 
 # argument.
  def prn(name)
    puts " #{name} = " + to_s
  end 

# Returns a real number, the significand of which has not more
# than n digits. Notice that there is also a function round
# which takes no argument and which returns an integer number.
# This function replaces function Float#round.
  def round(*arg)
    n = arg.size
    case n
    when 0
      (self + 0.5).floor # output is integer
    when 1
      m = arg[0].to_i
      x = frexp
      y = x[0].ldexp(m)
      (y + 0.5).floor.to_f.ldexp(x[1] - m)
    else
      fail "needs 0 or 1 arguments"
    end
  end
  
end # class Float
