=begin rdoc
ruby

Ulrich Mutze, www.ulrichutze.de

Started 2008-11-01 by modifying cpminterval.h and cpminterval.cpp.

This is free software. It comes without any warranty.
Please contact the author to report problems or to
ask questions.
=end 

require 'appmath_basics.rb'

module AppMath

=begin rdoc
Class of real closed intervals.
In most applications of intervals it is an advantage if single real
values can be considered as intervals of length zero. This is possible
only for closed intervals, and this is the rational for defining those
in preference to open intervals. 
With the attributes @low and @upp, the values x which belong to interval
are characterized by the formula
  @low <= x <= @upp .
Hence the interval is empty (void) iff @low > @upp.
=end

class Iv

  attr_reader :low, :upp

# Allowed are 0, 1, 2 arguments, which all have to be convertible to R.
# For no argument the empty interval is created. For one argument, an
# second argument 0 is understood. For two arguments, these are the
# boundaries of the interval and the order on input does not matter:
#   iv1 = Iv.new(2,3); iv2 = Iv.new(3,2) 
# are thus the same, non-empty, interval, which in mathematical notation
# would be
#   [2,3]
# Finally
#   iv3 = Iv.new
# is an empty interval.
  def initialize(*arg)
    n = arg.size
    case n
    when 0
      @low = R.c1
      @upp = R.c0 # notice: epmpty since @low > @upp
    when 1
      x = R.c arg[0] 
      zero = R.c0
      if x < zero
        @low = x
        @upp = zero
      else
        @upp = x
        @low = zero
      end
    when 2
      x = R.c arg[0]
      y = R.c arg[1]
      if x < y
        @low = x
        @upp = y
      else
        @low = y 
        @upp = x
      end
    else
        fail "Iv.new takes 0 or 1 or 2 arguments, but not " + n.to_s
    end # case n
  end  
  
=begin rdoc
  using max and min of an array for defining an interval.
  The values of the array components are then known to belong to
  to a closed interval. The smallest such interval is what
  his method returns.
=end
  def Iv.from_array(anArray)
    n = anArray.size
    case n
    when 0
      Iv.new
    when 1 
      a0 = anArray[0]
      Iv.new(a0)
    else
      x0 = anArray[0]
      x1 = x0
      anArray.each{ |x|
        if x < x0 
          x0 = x
        elsif x > x1
          x1 = x
        else
        end
      }
      Iv.new(x0,x1)
    end
  end
  
# Returns an ordered equidistant array of n numbers, where the first one 
# is @low and the last one is @upp. The positive integer n is made from
# the argument a in a way which depends on whether this argument is integer
# or not. In the first case it is taken as n and in the second case, the
# argment is taken as a proposal for the lattice spacing. The actual
# lattice spacing will be chosen as a fraction of size.
  def to_array(a)
    return nil if empty?
    if a.integer? # then we interpret the argument as an intended number
      # of lattice points
      n = a
    else # then we interprete the argument as an intended lattice spacing
      d = a.abs
      fail "can't build a lattice with spacing 0" if d.zero?
      n = (size/d).round + 1
    end
    fail "number of lattice points must be larger than 1" if n < 2
    res = Array.new
    res << @low
    if n > 2
      na = n - 2
      d = size / (n-1)
      x = @low
      for i in 1..na
        x += d
        res << x
      end
    end
    res << @upp
  end
 
# Returns true iff the interval is empty 
  def empty?; @low > @upp; end
  
# Returns the lower boundary if self is not empty
# and nil else.
  def inf
    return nil if empty?
    @low
  end
  
# Returns the upper boundary if self is not empty
# and nil else.
  def sup
    return nil if empty?
    @upp
  end  

# Returns the midpoint of the interval. 
  def center
    nil if empty?
    (@low + @upp) * R.i2
  end

# Returns the length of the interval, 0 for the empty one
  def size
    s = @upp - @low
    s >= R.c0 ? s : R.c0
  end
 
# Indicator function. Returns true if the point x belongs to self
# and false else.
  def ind(x)
    return false if empty?
    return false if x > @upp
    return false if x < @low
    true
  end
    
# Metrical indicator function. Returns the distance of x from the set
# self if x is outside of self. If it is inside, the return value is
# minus the distance to the complement of self (which is not an 
# interval but a well-defined set).
  def met_ind(x)
    return nil if empty?
    y = R.c x
    dc = (y - center).abs
    dc - size * R.i2
  end
  
# Returns a real number, which is self's lower end for p == 0.0, self's
# center for p == 0.5, and self's upper end for p == 1.0 .
  def put(p)
    inf + size * p
  end
 
# minimum closed interval that contains the union
# join or l.u.b. (lowest upper bound) in lattice terminology 
  def |(anIv)
    return anIv if empty?
    return self if anIv.empty?
    Iv.from_array [@low, @upp, anIv.low, anIv.upp]
  end
 
#shifting by a 
  def +(a); Iv.new(@low + a,@upp + a); end

# multiplying size by a, while preserving the center
  def *(a)
    Iv.new(center-(center-@low)*a,center+(@upp-center)*a)
  end
 
# section, intersection
# meet or g.l.b (greatest lower bound) in lattice terminology
  def &(anIv)
    return Iv.new if empty? || anIv.empty?
    return Iv.new if @upp < anIv.low || @low > anIv.upp
    amin = Basics.sup(@low, anIv.low)
    amax = Basics.inf(@upp, anIv.upp)
    Iv.new(amin, amax)
  end
  
  def to_s
    "Iv (" + @low.to_s + "," + @upp.to_s + ")"
  end

# The function creates a suitable axis sub-division for data ranging
# from @low to @upp. Let res be the return value of the function. Then
# res[0] is a proposal for the difference between adjacent axis tics
# and res[1] is an array of the values to which the proposed tics belong.
# Thus res[1].first <= @low and res[1].last >= @upp. All numbers are chosen
# such that they are simple when written down in normal scientific
# notation and the intention is to simulate the considerations that 
# determine the axis subdivision of reasonable manually created diagrams. 
# The argument of the function is a proposal for the number of tics to be 
# used. Values from 5 to 10 are reasonable. To have a simple logic, we
# simply enforce that the interval between tics is a simple number.
# The initial and the final number of the axis division is chosen as 
# an integer multiple of this inter-tic interval. 

  def axis_division(anPosInteger)
    fail "can't divide an empty interval" if empty?
    a = @low
    b = @upp
    n = anPosInteger.abs
    n += 1 if n == 0
    d = size/n
    d_ = Basics.cut(d) # this is the essential point
    fail "Zero division in function axis_division" if d_ == 0.0
    d_inv=d_.inv;
    epsilon = R.c 1e-6
    k_b=(b*d_inv - epsilon).ceil
      # without the epsilon correction it depends on roundoff 
      # errors whether b_ becomes too large
    k_a=(a*d_inv + epsilon).floor
      # without the epsilon correction it depends on roundoff 
      # errors whether a_ becomes too small
    b_=k_b * d_
    b__ = (k_b - 0.5) * d_
    a_=k_a * d_
    res=Array.new
    while a_ < b__
      res << a_
      a_ += d_
    end
    res << b_ # we know the last item exactly, and should
      # not spoil it by arithmetic errors
    [ d_ , res ]
  end

end # class Iv

end # AppMath
